@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2023 Paul Mattes.
+ * Copyright (c) 2014-2025 Paul Mattes.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -47,12 +47,13 @@
 #include "bind-opt.h"
 #include "json.h"
 #include "json_run.h"
-#include "lazya.h"
 #include "popups.h"
 #include "resources.h"
+#include "s3270_proto.h"
 #include "task.h"
 #include "toggles.h"
 #include "trace.h"
+#include "txa.h"
 #include "utils.h"
 #include "varbuf.h"
 
@@ -81,7 +82,6 @@ struct hio_listener {
 static hio_listener_t *global_listener = NULL;
 static llist_t listeners = LLIST_INIT(listeners);
 
-#define N_SESSIONS	32
 typedef struct {
     llist_t link;	/* list linkage */
     socket_t s;		/* socket */
@@ -147,6 +147,34 @@ hio_socket_close(session_t *session)
     Free(session);
 }
 
+/**
+ * Walk the sessions to find the one that matches the ID.
+ * If found, kill it.
+ *
+ * @param[in] id	I/O ID
+ */
+void
+hio_error_timeout(ioid_t id)
+{
+    session_t *session = NULL;
+    session_t *fatal_session = NULL;
+
+    vtrace("httpd deferred error timeout\n");
+    FOREACH_LLIST(&sessions, session, session_t *) {
+	if (httpd_waiting(session->dhandle, id)) {
+	    fatal_session = session;
+	    break;
+	}
+    } FOREACH_LLIST_END(&sessions, session, session_t *);
+    if (fatal_session == NULL) {
+	vtrace("httpd deferred error timeout: not found\n");
+	return;
+    }
+
+    httpd_close(fatal_session->dhandle, "deferred error timeout");
+    hio_socket_close(fatal_session);
+}
+
 /**
  * httpd timeout.
  *
@@ -217,7 +245,7 @@ hio_socket_input(iosrc_t fd, ioid_t id)
 	    if (socket_errno() == SE_EWOULDBLOCK) {
 		harmless = true;
 	    }
-	    ebuf = lazyaf("recv error: %s", socket_errtext());
+	    ebuf = txAsprintf("recv error: %s", socket_errtext());
 	    vtrace("httpd %s%s\n", ebuf, harmless? " (harmless)": "");
 	} else {
 	    ebuf = "session EOF";
@@ -285,11 +313,6 @@ hio_connection(iosrc_t fd, ioid_t id)
 		(socket_errno() == SE_EWOULDBLOCK)? " (harmless)": "");
 	return;
     }
-    if (l->n_sessions >= N_SESSIONS) {
-	vtrace("Too many connections.\n");
-	SOCK_CLOSE(t);
-	return;
-    }
 
 #if !defined(_WIN32) /*[*/
     fcntl(t, F_SETFD, 1);
@@ -319,13 +342,13 @@ hio_connection(iosrc_t fd, ioid_t id)
 #endif /*]*/
     if (sa.sa.sa_family == AF_INET) {
 	session->dhandle = httpd_new(session,
-		lazyaf("%s:%u",
+		txAsprintf("%s:%u",
 		    inet_ntop(AF_INET, &sa.sin.sin_addr, hostbuf,
 			sizeof(hostbuf)),
 		    ntohs(sa.sin.sin_port)));
     } else if (sa.sa.sa_family == AF_INET6) {
 	session->dhandle = httpd_new(session,
-		lazyaf("%s:%u",
+		txAsprintf("%s:%u",
 		    inet_ntop(AF_INET6, &sa.sin6.sin6_addr, hostbuf,
 			sizeof(hostbuf)),
 		    ntohs(sa.sin6.sin6_port)));
@@ -357,7 +380,9 @@ hio_connection(iosrc_t fd, ioid_t id)
 hio_listener_t *
 hio_init_x(struct sockaddr *sa, socklen_t sa_len)
 {
+#if !defined(_WIN32) /*[*/
     int on = 1;
+#endif /*]*/
     char hostbuf[128];
 
     hio_listener_t *l = Calloc(sizeof(hio_listener_t), 1);
@@ -373,6 +398,7 @@ hio_init_x(struct sockaddr *sa, socklen_t sa_len)
 	popup_an_error("httpd socket: %s", socket_errtext());
 	goto fail;
     }
+#if !defined(_WIN32) /*[*/
     if (setsockopt(l->listen_s, SOL_SOCKET, SO_REUSEADDR, (char *)&on,
 		sizeof(on)) < 0) {
 	popup_an_error("httpd setsockopt: %s", socket_errtext());
@@ -380,6 +406,7 @@ hio_init_x(struct sockaddr *sa, socklen_t sa_len)
 	l->listen_s = INVALID_SOCKET;
 	goto fail;
     }
+#endif /*]*/
     if (bind(l->listen_s, sa, sa_len) < 0) {
 	popup_an_error("httpd bind: %s", socket_errtext());
 	SOCK_CLOSE(l->listen_s);
@@ -427,14 +454,14 @@ hio_init_x(struct sockaddr *sa, socklen_t sa_len)
     if (sa->sa_family == AF_INET) {
 	struct sockaddr_in *sin = (struct sockaddr_in *)sa;
 
-	l->desc = xs_buffer("%s:%u", inet_ntop(sa->sa_family,
+	l->desc = Asprintf("%s:%u", inet_ntop(sa->sa_family,
 		    &sin->sin_addr, hostbuf, sizeof(hostbuf)),
 		ntohs(sin->sin_port));
 	vtrace("Listening for HTTP on %s\n", l->desc);
     } else if (sa->sa_family == AF_INET6) {
 	struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *)sa;
 
-	l->desc = xs_buffer("[%s]:%u", inet_ntop(sa->sa_family,
+	l->desc = Asprintf("[%s]:%u", inet_ntop(sa->sa_family,
 		&sin6->sin6_addr, hostbuf, sizeof(hostbuf)),
 	    ntohs(sin6->sin6_port));
 	vtrace("Listening for HTTP on %s\n", l->desc);
@@ -546,6 +573,11 @@ hio_data(task_cbh handle, const char *buf, size_t len, bool success)
 {
     session_t *s = handle;
 
+    /* Remove trailing newlines. */
+    while (len > 0 && buf[len - 1] == '\n') {
+        len--;
+     }
+
     if (s->pending.return_content_type == CT_HTML) {
 	size_t i;
 	char c;
@@ -571,24 +603,39 @@ hio_data(task_cbh handle, const char *buf, size_t len, bool success)
 		break;
 	    }
 	}
+	vb_appends(&s->pending.result, "\n");
     } else if (s->pending.return_content_type == CT_JSON) {
 	json_t *result_array;
+	json_t *err_array;
+	char *bcopy;
+	char *bnext;
+	char *newline;
 
 	if (s->pending.jresult == NULL) {
 	    s->pending.jresult = json_object();
 	    result_array = json_array();
-	    json_object_set(s->pending.jresult, "result", NT, result_array);
+	    json_object_set(s->pending.jresult, JRET_RESULT, NT, result_array);
+	    err_array = json_array();
+	    json_object_set(s->pending.jresult, JRET_RESULT_ERR, NT, err_array);
 	} else {
-	    assert(json_object_member(s->pending.jresult, "result", NT,
-			&result_array));
+	    assert(json_object_member(s->pending.jresult, JRET_RESULT, NT, &result_array));
+	    assert(json_object_member(s->pending.jresult, JRET_RESULT_ERR, NT, &err_array));
 	}
 
-	json_array_append(result_array, json_string(buf, len));
+	bnext = bcopy = Asprintf("%.*s", (int)len, buf);
+	while ((newline = strchr(bnext, '\n')) != NULL) {
+	    json_array_append(result_array, json_string(bnext, newline - bnext));
+	    json_array_append(err_array, json_boolean(!success));
+	    bnext = newline + 1;
+	}
+	json_array_append(result_array, json_string(bnext, strlen(bnext)));
+	json_array_append(err_array, json_boolean(!success));
+	Free(bcopy);
     } else {
 	/* Plain text. */
 	vb_append(&s->pending.result, buf, len);
+	vb_appends(&s->pending.result, "\n");
     }
-    vb_appends(&s->pending.result, "\n");
 }
 
 /**
@@ -670,11 +717,11 @@ hio_verb(void *dhandle)
 /**
  * Send a command to x3270.
  *
- * @param[in] cmd       command to send, not including the newline and
- * @param[in] callback  callback function for completion
- * @param[in] handle    handle to pass to completion callback function
- * @param[in] request_content_type How to process input
- * @param[in] return_content_type How to return content
+ * @param[in] cmd	command to send, not including the newline and
+ * @param[in] callback	callback function for completion
+ * @param[in] dhandle	handle to pass to completion callback function
+ * @param[in] request_content_type	How to process input
+ * @param[in] return_content_type	How to return content
  * @param[out] errmsg	error message, if SENDTO_INVALID returned
  *
  * @return sendto_t
@@ -782,7 +829,6 @@ hio_async_done(void *dhandle, httpd_status_t rv)
  *
  * @param[in] name	Name of toggle
  * @param[in] value	Toggle value
- * @param[out] canonical_value	Returned canonical form of value
  * @param[in] flags	Set() flags
  * @param[in] ia	Cause
  *
