@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2021-2022 Paul Mattes.
+ * Copyright (c) 2021-2024 Paul Mattes.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -349,7 +349,7 @@ ucs4streq(ucs4_t *a, ucs4_t *b)
 /**
  * Format an error message that ends with a Unicode character.
  * @param[in] text	Body of message
- * @param[in] ucs4	Unicode character
+ * @param[in] u		Unicode character
  *
  * @returns Formatted message
  */
@@ -357,8 +357,8 @@ static char *
 format_uerror(const char *text, ucs4_t u)
 {
     return (u < 0xff && isprint((int)u))?
-	xs_buffer("%s '%c'", text, u):
-	xs_buffer("%s U+%04x", text, u);
+	Asprintf("%s '%c'", text, u):
+	Asprintf("%s U+%04x", text, u);
 }
 
 /**
@@ -371,7 +371,7 @@ format_uerror(const char *text, ucs4_t u)
  * @param[out] result	Result if successful
  * @param[out] error	Error if not successful
  * @param[out] stop_token Character after this element
- * @param[out] r_any	True if something parsed
+ * @param[out] any	True if something parsed
  * @returns error code
  */
 static json_errcode_t
@@ -922,13 +922,13 @@ json_write_indent(const json_t *json, unsigned options, int indent)
     case JT_BOOLEAN:
 	return NewString(json_boolean_value(json)? "true": "false");
     case JT_INTEGER:
-	return xs_buffer("%"JSON_INT_PRINT, json_integer_value(json));
+	return Asprintf("%"JSON_INT_PRINT, json_integer_value(json));
     case JT_DOUBLE:
-	return xs_buffer("%g", json_double_value(json));
+	return Asprintf("%g", json_double_value(json));
     case JT_STRING:
 	v = json_string_value(json, &len);
 	s = json_expand_string(v, len, options);
-	t = xs_buffer("\"%s\"", s);
+	t = Asprintf("\"%s\"", s);
 	Free(s);
 	return t;
     case JT_OBJECT:
@@ -1251,7 +1251,7 @@ json_object(void)
  * @returns array
  */
 json_t *
-json_array()
+json_array(void)
 {
     json_t *j = Calloc(1, sizeof(json_t));
 
