@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-2022 Paul Mattes.
+ * Copyright (c) 1993-2025 Paul Mattes.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -50,15 +50,16 @@
 #include "json.h"
 #include "json_run.h"
 #include "kybd.h"
-#include "lazya.h"
 #include "names.h"
 #include "peerscript.h"
 #include "popups.h"
 #include "s3270_proto.h"
+#include "s3common.h"
 #include "source.h"
 #include "task.h"
 #include "telnet_core.h"
 #include "trace.h"
+#include "txa.h"
 #include "utils.h"
 #include "varbuf.h"
 #include "w3misc.h"
@@ -70,6 +71,8 @@ static bool peer_done(task_cbh handle, bool success, bool abort);
 static void peer_closescript(task_cbh handle);
 static void peer_setflags(task_cbh handle, unsigned flags);
 static unsigned peer_getflags(task_cbh handle);
+static void peer_setxflags(task_cbh handle, unsigned flags);
+static unsigned peer_getxflags(task_cbh handle);
 
 static void peer_setir(task_cbh handle, void *irhandle);
 static void *peer_getir(task_cbh handle);
@@ -90,7 +93,7 @@ static irv_t peer_irv = {
 static tcb_t peer_cb = {
     "s3sock",
     IA_SCRIPT,
-    CB_NEW_TASKQ | CB_PEER,
+    CB_NEW_TASKQ | CB_PEER | CB_NEEDCOOKIE,
     peer_data,
     peer_done,
     NULL,
@@ -99,14 +102,16 @@ static tcb_t peer_cb = {
     peer_getflags,
     &peer_irv,
     NULL,
-    peer_reqinput
+    peer_reqinput,
+    peer_setxflags,
+    peer_getxflags,
 };
 
 /* Callback block for an interactive peer. */
 static tcb_t interactive_cb = {
     "s3sock",
     IA_COMMAND,
-    CB_NEW_TASKQ | CB_PEER,
+    CB_NEW_TASKQ | CB_PEER | CB_NEEDCOOKIE,
     peer_data,
     peer_done,
     NULL,
@@ -115,13 +120,16 @@ static tcb_t interactive_cb = {
     peer_getflags,
     &peer_irv,
     NULL,
-    peer_reqinput
+    peer_reqinput,
+    peer_setxflags,
+    peer_getxflags,
 };
 
 /* Peer script context. */
 typedef struct {
     llist_t llist;	/* list linkage */
     socket_t socket;	/* socket */
+    char *desc;		/* description */
 #if defined(_WIN32) /*[*/
     HANDLE event;	/* event */
 #endif /*]*/
@@ -133,9 +141,9 @@ typedef struct {
     bool enabled;	/* is this peer enabled? */
     char *name;		/* task name */
     unsigned capabilities; /* self-reported capabilities */
+    unsigned xflags;	/* extended flags */
     void *irhandle;	/* input request handle */
     task_cb_ir_state_t ir_state; /* named input request state */
-    bool json_mode;	/* true if in JSON mode */
     json_t *json_result; /* pending JSON result */
 } peer_t;
 static llist_t peer_scripts = LLIST_INIT(peer_scripts);
@@ -176,6 +184,7 @@ close_peer(peer_t *p)
 	RemoveInput(p->id);
 	p->id = NULL_IOID;
     }
+    Replace(p->desc, NULL);
     Replace(p->buf, NULL);
     Replace(p->name, NULL);
 
@@ -213,8 +222,7 @@ do_push(peer_t *p, const char *buf, size_t len)
     }
 
     /* Try JSON parsing. */
-    if (!(p->capabilities & CBF_INTERACTIVE) &&
-	    (*s == '{' || *s == '[' || *s == '"')) {
+    if (!(p->capabilities & CBF_INTERACTIVE) && (*s == '{' || *s == '[' || *s == '"')) {
 	cmd_t **cmds;
 	char *single;
 	char *errmsg;
@@ -223,7 +231,7 @@ do_push(peer_t *p, const char *buf, size_t len)
 	ret = hjson_parse(s, len, &cmds, &single, &errmsg);
 	if (ret == HJ_OK) {
 	    /* Good JSON. */
-	    p->json_mode = true;
+	    p->json_result = s3json_init();
 	    if (cmds != NULL) {
 		name = push_cb_split(cmds, tcb, (task_cbh)p);
 	    } else {
@@ -235,16 +243,18 @@ do_push(peer_t *p, const char *buf, size_t len)
 	    return false;
 	} else {
 	    /* Bad JSON. */
-	    char *fail = xs_buffer(AnFail "(\"%s\")", errmsg);
+	    char *fail = Asprintf(AnFail "(\"%s\")", errmsg);
 
 	    /* Answer in JSON only if successfully parsed. */
-	    p->json_mode = (ret != HJ_BAD_SYNTAX);
+	    if (ret != HJ_BAD_SYNTAX) {
+		p->json_result = s3json_init();
+	    }
 	    Free(errmsg);
 	    name = push_cb(fail, strlen(fail), tcb, (task_cbh)p);
 	    Free(fail);
 	}
     } else {
-	p->json_mode = false;
+	json_free(p->json_result);
 	name = push_cb(s, len, tcb, (task_cbh)p);
     }
     Replace(p->name, NewString(name));
@@ -302,7 +312,7 @@ run_next(peer_t *p)
 }
 
 /**
- * Read the next command from a peer socket (Unix version).
+ * Read the next command from a peer socket.
  * @param[in] fd	File descriptor
  * @param[in] id	I/O identifier
  */
@@ -332,18 +342,17 @@ peer_input(iosrc_t fd _is_unused, ioid_t id)
 #if defined(_WIN32) /*[*/
 	if (GetLastError() != WSAECONNRESET) {
 	    /* Windows does this habitually. */
-	    popup_an_error("s3sock recv: %s",
-		    win32_strerror(GetLastError()));
+	    vtrace("s3sock %s recv: %s\n", p->desc, win32_strerror(GetLastError()));
 	}
 #else /*][*/
-	popup_an_errno(errno, "s3sock read");
+	vtrace("s3sock %s recv: %s\n", p->desc, strerror(errno));
 #endif /*]*/
 	close_peer(p);
 	return;
     }
-    vtrace("Input for s3sock complete, nr=%d\n", (int)nr);
+    vtrace("Input for s3sock %s complete, nr=%d\n", p->desc, (int)nr);
     if (nr == 0) {
-	vtrace("s3sock EOF\n");
+	vtrace("s3sock %s EOF\n", p->desc);
 	close_peer(p);
 	return;
     }
@@ -375,6 +384,31 @@ peer_input(iosrc_t fd _is_unused, ioid_t id)
     }
 }
 
+/**
+ * Send data on a socket and check the result.
+ *
+ * @param[in] s		Socket
+ * @param[in] data	Data to send
+ * @param[in] len	Length
+ * @param[in] sender	Sending function
+ */
+static void
+check_send(socket_t s, const char *data, size_t len, const char *sender)
+{
+    ssize_t ns = send(s, data, (int)len, 0);
+    if (ns != (ssize_t)len) {
+	if (ns < 0) {
+#if !defined(_WIN32) /*[*/
+	    vtrace("%s send: %s\n", sender, strerror(errno));
+#else /*][*/
+	    vtrace("%s send: %s\n", sender, win32_strerror(GetLastError()));
+#endif/*]*/
+	} else {
+	    vtrace("%s: short send\n", sender);
+	}
+    }
+}
+
 /**
  * Callback for data returned to peer socket command.
  *
@@ -387,8 +421,7 @@ static void
 peer_data(task_cbh handle, const char *buf, size_t len, bool success)
 {
     peer_t *p = (peer_t *)handle;
-    char *s;
-    ssize_t ns;
+    char *cooked;
     static bool recursing = false;
 
     if (recursing) {
@@ -396,24 +429,10 @@ peer_data(task_cbh handle, const char *buf, size_t len, bool success)
     }
     recursing = true;
 
-    if (p->json_mode) {
-	json_t *result_array;
-
-	if (p->json_result == NULL) {
-            p->json_result = json_object();
-            result_array = json_array();
-            json_object_set(p->json_result, "result", NT, result_array);
-        } else {
-            assert(json_object_member(p->json_result, "result", NT,
-                        &result_array));
-        }
-	json_array_append(result_array, json_string(buf, len));
-    } else {
-	s = lazyaf(DATA_PREFIX "%.*s\n", (int)len, buf);
-	ns = send(p->socket, s, (int)strlen(s), 0);
-	if (ns < 0) {
-	    popup_a_sockerr("s3sock send");
-	}
+    s3data(buf, len, success, p->capabilities, p->json_result, NULL, &cooked);
+    if (cooked != NULL) {
+	check_send(p->socket, cooked, strlen(cooked), "peer_data");
+	Free(cooked);
     }
 
     recursing = false;
@@ -431,19 +450,17 @@ static void
 peer_reqinput(task_cbh handle, const char *buf, size_t len, bool echo)
 {
     peer_t *p = (peer_t *)handle;
-    char *s = lazyaf("%s%.*s\n", echo? INPUT_PREFIX: PWINPUT_PREFIX, (int)len,
-	    buf);
-    ssize_t ns;
+    char *s;
     static bool recursing = false;
 
     if (recursing) {
 	return;
     }
     recursing = true;
-    ns = send(p->socket, s, (int)strlen(s), 0);
-    if (ns < 0) {
-	popup_a_sockerr("s3sock send");
-    }
+
+    s = Asprintf("%s%.*s\n", echo? INPUT_PREFIX: PWINPUT_PREFIX, (int)len, buf);
+    check_send(p->socket, s, strlen(s), "peer_reqinput");
+    Free(s);
     recursing = false;
 }
 
@@ -460,33 +477,12 @@ static bool
 peer_done(task_cbh handle, bool success, bool abort)
 {
     peer_t *p = (peer_t *)handle;
-    char *prompt = task_cb_prompt(handle);
-    char *s;
+    char *out;
     bool new_child = false;
 
-    if (p->json_mode) {
-	char *w;
-	/* Print the result. */
-
-        if (p->json_result == NULL) {
-            p->json_result = json_object();
-            json_object_set(p->json_result, "result", NT, json_array());
-        }
-        json_object_set(p->json_result, "success", NT, json_boolean(success));
-        json_object_set(p->json_result, "status", NT, json_string(prompt, NT));
-        s = lazyaf("%s\n", w = json_write_o(p->json_result, JW_ONE_LINE));
-        Free(w);
-        json_free(p->json_result);
-	vtrace("JSON output for %s: '%s/%s'\n", p->name, prompt,
-		success? PROMPT_OK: PROMPT_ERROR);
-	send(p->socket, s, (int)strlen(s), 0);
-    } else {
-	/* Print the prompt. */
-	vtrace("Output for %s: '%s/%s'\n", p->name, prompt,
-		success? PROMPT_OK: PROMPT_ERROR);
-	s = lazyaf("%s\n%s\n", prompt, success? PROMPT_OK: PROMPT_ERROR);
-	send(p->socket, s, (int)strlen(s), 0);
-    }
+    s3done(handle, success, &p->json_result, &out);
+    check_send(p->socket, out, strlen(out), "peer_done");
+    Free(out);
 
     if (abort || !p->enabled) {
 	close_peer(p);
@@ -564,6 +560,34 @@ peer_setir(task_cbh handle, void *irhandle)
     p->irhandle = irhandle;
 }
 
+/**
+ * Get extended flags.
+ *
+ * @param[in] handle	Peer context
+ * @returns flags
+ */
+static unsigned
+peer_getxflags(task_cbh handle)
+{
+    peer_t *p = (peer_t *)handle;
+
+    return p->xflags;
+}
+
+/**
+ * Set extended flags.
+ *
+ * @param[in] handle	Peer context
+ * @param[in] flags	Flags
+ */
+static void
+peer_setxflags(task_cbh handle, unsigned flags)
+{
+    peer_t *p = (peer_t *)handle;
+
+    p->xflags = flags;
+}
+
 /**
  * Get the pending input request.
  *
@@ -629,6 +653,7 @@ peer_connection(iosrc_t fd _is_unused, ioid_t id)
     char hostbuf[128];
     peer_listen_t listener;
     bool found = false;
+    const char *desc;
 
     FOREACH_LLIST(&peer_listeners, listener, peer_listen_t) {
 	if (listener->id == id) {
@@ -641,26 +666,29 @@ peer_connection(iosrc_t fd _is_unused, ioid_t id)
     accept_fd = accept(listener->socket, &sa.sa, &len);
     if (accept_fd != INVALID_SOCKET) {
 	if (sa.sa.sa_family == AF_INET) {
-	    vtrace("New script socket connection from %s:%u\n",
-		    inet_ntop(AF_INET, &sa.sin.sin_addr, hostbuf,
-			sizeof(hostbuf)), ntohs(sa.sin.sin_port));
+	    desc = txAsprintf("%s:%u", inet_ntop(AF_INET, &sa.sin.sin_addr, hostbuf, sizeof(hostbuf)),
+		    ntohs(sa.sin.sin_port));
 	} else if (sa.sa.sa_family == AF_INET6) {
-	    vtrace("New script socket connection from %s:%u\n",
-		    inet_ntop(AF_INET6, &sa.sin6.sin6_addr, hostbuf,
-			sizeof(hostbuf)), ntohs(sa.sin6.sin6_port));
+	    desc = txAsprintf("[%s]:%u", inet_ntop(AF_INET6, &sa.sin6.sin6_addr, hostbuf, sizeof(hostbuf)),
+		    ntohs(sa.sin6.sin6_port));
 	}
 #if !defined(_WIN32) /*[*/
 	else if (sa.sa.sa_family == AF_UNIX) {
-	    vtrace("New Unix-domain script socket connection");
+	    desc = "Unix-domain socket";
 	}
 #endif /*]*/
 	else {
-	    vtrace("New script socket connection from ???\n");
+	    desc = "???";
 	}
+	vtrace("New script socket connection from %s\n", desc);
     }
 
     if (accept_fd == INVALID_SOCKET) {
-	popup_an_errno(errno, "script socket accept");
+#if !defined(_WIN32) /*[*/
+	vtrace("s3sock accept: %s\n", strerror(errno));
+#else /*][*/
+	vtrace("s3sock accept: %s\n", win32_strerror(GetLastError()));
+#endif /*]*/
 	return;
     }
 
@@ -686,7 +714,7 @@ peer_connection(iosrc_t fd _is_unused, ioid_t id)
     }
 
     /* Allocate the peer state and remember it. */
-    peer_accepted(accept_fd, listener);
+    peer_accepted(accept_fd, listener, desc);
 }
 
 /**
@@ -694,9 +722,10 @@ peer_connection(iosrc_t fd _is_unused, ioid_t id)
  *
  * @param[in] s		Socket
  * @param[in] listener	Listener, or null
+ * @param[in] desc	Peer description
  */
 void
-peer_accepted(socket_t s, void *listener)
+peer_accepted(socket_t s, void *listener, const char *desc)
 {
     peer_t *p = (peer_t *)Calloc(1, sizeof(peer_t));
 #if defined(_WIN32) /*[*/
@@ -720,6 +749,7 @@ peer_accepted(socket_t s, void *listener)
     llist_init(&p->llist);
     p->listener = listener;
     p->socket = s;
+    p->desc = NewString(desc);
 #if defined(_WIN32) /*[*/
     p->event = event;
     p->id = AddInput(p->event, peer_input);
@@ -747,7 +777,9 @@ peer_init(struct sockaddr *sa, socklen_t sa_len, peer_listen_mode mode)
 {
     peer_listen_t listener;
     char hostbuf[128];
+#if !defined(_WIN32) /*[*/
     int on = 1;
+#endif /*]*/
 
     /* Create the listening socket. */
     listener = (peer_listen_t)Calloc(sizeof(struct _peer_listen), 1);
@@ -774,16 +806,13 @@ peer_init(struct sockaddr *sa, socklen_t sa_len, peer_listen_mode mode)
 	goto fail;
     }
 
+#if !defined(_WIN32) /*[*/
     if (setsockopt(listener->socket, SOL_SOCKET, SO_REUSEADDR,
 		(char *)&on, sizeof(on)) < 0) {
-#if !defined(_WIN32) /*[*/
 	popup_an_errno(errno, "script setsockopt(SO_REUSEADDR)");
-#else /*][*/
-	popup_an_error("script setsockopt(SO_REUSEADDR): %s",
-		win32_strerror(GetLastError()));
-#endif /*]*/
 	goto fail;
     }
+#endif /*]*/
 
     if (bind(listener->socket, sa, sa_len) < 0) {
 #if !defined(_WIN32) /*[*/
@@ -839,14 +868,14 @@ peer_init(struct sockaddr *sa, socklen_t sa_len, peer_listen_mode mode)
     if (sa->sa_family == AF_INET) {
 	struct sockaddr_in *sin = (struct sockaddr_in *)sa;
 
-	listener->desc = xs_buffer("%s:%u", inet_ntop(sa->sa_family,
+	listener->desc = Asprintf("%s:%u", inet_ntop(sa->sa_family,
 		    &sin->sin_addr, hostbuf, sizeof(hostbuf)),
 		ntohs(sin->sin_port));
 	vtrace("Listening for s3sock scripts on %s\n", listener->desc);
     } else if (sa->sa_family == AF_INET6) {
 	struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *)sa;
 
-	listener->desc = xs_buffer("[%s]:%u", inet_ntop(sa->sa_family,
+	listener->desc = Asprintf("[%s]:%u", inet_ntop(sa->sa_family,
 		    &sin6->sin6_addr, hostbuf, sizeof(hostbuf)),
 		ntohs(sin6->sin6_port));
 	vtrace("Listening for s3sock scripts on %s\n", listener->desc);
