@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 #
-# Copyright (c) 2021-2022 Paul Mattes.
+# Copyright (c) 2021-2025 Paul Mattes.
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -27,25 +27,26 @@
 #
 # s3270 code page tests
 
-import unittest
 from subprocess import Popen, PIPE, DEVNULL
-import requests
-import Common.Test.playback as playback
-import Common.Test.cti as cti
+import unittest
 
-class TestS3270CodePage(cti.cti):
+from Common.Test.cti import *
+from Common.Test.playback import playback
+
+@requests_timeout
+class TestS3270CodePage(cti):
 
     # s3270 SBCS code page test.
     def test_s3270_sbcs_code_page(self):
 
         # Start playback.
-        pport, ts = cti.unused_port()
-        with playback.playback(self, 's3270/Test/all_chars.trc', port=pport) as p:
+        pport, ts = unused_port()
+        with playback(self, 's3270/Test/all_chars.trc', port=pport) as p:
             ts.close()
 
             # Start s3270.
-            sport, ts = cti.unused_port()
-            s3270 = Popen(cti.vgwrap(['s3270', '-httpd', str(sport), '-utf8', f'127.0.0.1:{pport}']))
+            sport, ts = unused_port()
+            s3270 = Popen(vgwrap(['s3270', '-httpd', str(sport), '-utf8', f'127.0.0.1:{pport}']))
             self.children.append(s3270)
             self.check_listen(sport)
             ts.close()
@@ -172,12 +173,12 @@ class TestS3270CodePage(cti.cti):
 
             # Check the SBCS code pages.
             for cp in cp_all_map.keys():
-                requests.get(f'http://127.0.0.1:{sport}/3270/rest/json/Set(codePage,{cp})')
-                r = requests.get(f'http://127.0.0.1:{sport}/3270/rest/json/Ascii1(2,2,12,16)')
+                self.get(f'http://127.0.0.1:{sport}/3270/rest/json/Set(codePage,{cp})')
+                r = self.get(f'http://127.0.0.1:{sport}/3270/rest/json/Ascii1(2,2,12,16)')
                 self.assertEqual(cp_all_map[cp], r.json()['result'])
 
         # Wait for the process to exit successfully.
-        requests.get(f'http://127.0.0.1:{sport}/3270/rest/json/Quit()')
+        self.get(f'http://127.0.0.1:{sport}/3270/rest/json/Quit()')
         self.vgwait(s3270)
 
     # s3270 APL code page test.
@@ -199,13 +200,13 @@ class TestS3270CodePage(cti.cti):
         ]
 
         # Start playback.
-        pport, ts = cti.unused_port()
-        with playback.playback(self, 's3270/Test/apl.trc', port=pport) as p:
+        pport, ts = unused_port()
+        with playback(self, 's3270/Test/apl.trc', port=pport) as p:
             ts.close()
 
             # Start s3270.
-            sport, ts = cti.unused_port()
-            s3270 = Popen(cti.vgwrap(['s3270', '-httpd', str(sport), '-utf8', f'127.0.0.1:{pport}']))
+            sport, ts = unused_port()
+            s3270 = Popen(vgwrap(['s3270', '-httpd', str(sport), '-utf8', f'127.0.0.1:{pport}']))
             self.children.append(s3270)
             self.check_listen(sport)
             ts.close()
@@ -214,11 +215,11 @@ class TestS3270CodePage(cti.cti):
             p.send_records(1)
 
         # Check the SBCS code pages.
-        r = requests.get(f'http://127.0.0.1:{sport}/3270/rest/json/Ascii1(2,2,12,16)')
+        r = self.get(f'http://127.0.0.1:{sport}/3270/rest/json/Ascii1(2,2,12,16)')
         self.assertEqual(apl_chars, r.json()['result'])
         
         # Wait for the process to exit successfully.
-        requests.get(f'http://127.0.0.1:{sport}/3270/rest/json/Quit()')
+        self.get(f'http://127.0.0.1:{sport}/3270/rest/json/Quit()')
         self.vgwait(s3270)
 
     # s3270 simplified Chinese test.
@@ -231,13 +232,13 @@ class TestS3270CodePage(cti.cti):
         ]
 
         # Start playback.
-        pport, ts = cti.unused_port()
-        with playback.playback(self, 's3270/Test/935.trc', port=pport) as p:
+        pport, ts = unused_port()
+        with playback(self, 's3270/Test/935.trc', port=pport) as p:
             ts.close()
 
             # Start s3270.
-            sport, ts = cti.unused_port()
-            s3270 = Popen(cti.vgwrap(['s3270', '-httpd', str(sport), '-utf8', '-codepage', codePage, f'127.0.0.1:{pport}']))
+            sport, ts = unused_port()
+            s3270 = Popen(vgwrap(['s3270', '-httpd', str(sport), '-utf8', '-codepage', codePage, f'127.0.0.1:{pport}']))
             self.children.append(s3270)
             self.check_listen(sport)
             ts.close()
@@ -246,11 +247,11 @@ class TestS3270CodePage(cti.cti):
             p.send_records(1)
 
             # Check the DBCS output.
-            r = requests.get(f'http://127.0.0.1:{sport}/3270/rest/json/Ascii1(2,1,3,80)')
+            r = self.get(f'http://127.0.0.1:{sport}/3270/rest/json/Ascii1(2,1,3,80)')
             self.assertEqual(expect_chinese, r.json()['result'])
         
         # Wait for the process to exit successfully.
-        requests.get(f'http://127.0.0.1:{sport}/3270/rest/json/Quit()')
+        self.get(f'http://127.0.0.1:{sport}/3270/rest/json/Quit()')
         self.vgwait(s3270)
 
     # Simplified Chinese tests.
@@ -264,17 +265,17 @@ class TestS3270CodePage(cti.cti):
 
         expect_chinese = [
             '   辦理中止委託轉帳代繳，須提供原代繳帳號，申辦資料如有疑義或不明，經本公司聯繫補',
-            '正資料，請配合提供補正；不願提供者不予受理。                                   '
+            ' 正資料，請配合提供補正；不願提供者不予受理。                                   '
         ]
 
         # Start playback.
-        pport, ts = cti.unused_port()
-        with playback.playback(self, 's3270/Test/937.trc', port=pport) as p:
+        pport, ts = unused_port()
+        with playback(self, 's3270/Test/937.trc', port=pport) as p:
             ts.close()
 
             # Start s3270.
-            sport, ts = cti.unused_port()
-            s3270 = Popen(cti.vgwrap(['s3270', '-httpd', str(sport), '-utf8', '-codepage', 'cp937', f'127.0.0.1:{pport}']))
+            sport, ts = unused_port()
+            s3270 = Popen(vgwrap(['s3270', '-httpd', str(sport), '-utf8', '-codepage', 'cp937', f'127.0.0.1:{pport}']))
             self.children.append(s3270)
             self.check_listen(sport)
             ts.close()
@@ -283,11 +284,11 @@ class TestS3270CodePage(cti.cti):
             p.send_records(1)
 
             # Check the DBCS output.
-            r = requests.get(f'http://127.0.0.1:{sport}/3270/rest/json/Ascii1(2,1,2,80)')
+            r = self.get(f'http://127.0.0.1:{sport}/3270/rest/json/Ascii1(2,1,2,80)')
             self.assertEqual(expect_chinese, r.json()['result'])
         
         # Wait for the process to exit successfully.
-        requests.get(f'http://127.0.0.1:{sport}/3270/rest/json/Quit()')
+        self.get(f'http://127.0.0.1:{sport}/3270/rest/json/Quit()')
         self.vgwait(s3270)
 
     # s3270 Japanese DBCS test.
@@ -297,21 +298,29 @@ class TestS3270CodePage(cti.cti):
             'cp930': [
                 ' Jｧﾆｧﾄｵﾍｵ ﾎｵﾍﾎ ｶﾅﾈ ｳﾅｴｵ ﾆｧｷｵ 930                                                ',
                 '   高度な音声信号処理を行う高精度ボイスピックアップテクノロジーにより、高い通話品',
-                '質を実現。 AI による機械学習アルゴリズムで実現されたノイズリダクションシステムが'],
+                ' 質を実現。 AI による機械学習アルゴリズムで実現されたノイズリダクションシステムが'],
             'cp939': [
                 ' Japanese test for code page 930                                                ',
                 '   高度な音声信号処理を行う高精度ボイスピックアップテクノロジーにより、高い通話品',
-                '質を実現。 AI による機械学習アルゴリズムで実現されたノイズリダクションシステムが']
+                ' 質を実現。 AI による機械学習アルゴリズムで実現されたノイズリダクションシステムが'],
+            'cp1390': [
+                ' Jｧﾆｧﾄｵﾍｵ ﾎｵﾍﾎ ｶﾅﾈ ｳﾅｴｵ ﾆｧｷｵ 930                                                ',
+                '   高度な音声信号処理を行う高精度ボイスピッ アップテ ノロジーにより、高い通話品',
+                ' 質を実現。 AI による機械学習アルゴリズムで実現されたノイズリダ ションシステムが'],
+            'cp1399': [
+                ' Japanese test for code page 930                                                ',
+                '   高度な音声信号処理を行う高精度ボイスピックアップテクノロジーにより、高い通話品',
+                ' 質を実現。 AI による機械学習アルゴリズムで実現されたノイズリダクションシステムが']
         }
 
         # Start playback.
-        pport, ts = cti.unused_port()
-        with playback.playback(self, 's3270/Test/930.trc', port=pport) as p:
+        pport, ts = unused_port()
+        with playback(self, 's3270/Test/930.trc', port=pport) as p:
             ts.close()
 
             # Start s3270.
-            sport, ts = cti.unused_port()
-            s3270 = Popen(cti.vgwrap(['s3270', '-httpd', str(sport), '-utf8', '-codepage', codePage, f'127.0.0.1:{pport}']))
+            sport, ts = unused_port()
+            s3270 = Popen(vgwrap(['s3270', '-httpd', str(sport), '-utf8', '-codepage', codePage, f'127.0.0.1:{pport}']))
             self.children.append(s3270)
             self.check_listen(sport)
             ts.close()
@@ -320,17 +329,117 @@ class TestS3270CodePage(cti.cti):
             p.send_records(1)
 
             # Check the DBCS output.
-            r = requests.get(f'http://127.0.0.1:{sport}/3270/rest/json/Ascii1(1,1,3,80)')
+            r = self.get(f'http://127.0.0.1:{sport}/3270/rest/json/Ascii1(1,1,3,80)')
             self.assertEqual(expect_japanese[codePage], r.json()['result'])
         
         # Wait for the process to exit successfully.
-        requests.get(f'http://127.0.0.1:{sport}/3270/rest/json/Quit()')
+        self.get(f'http://127.0.0.1:{sport}/3270/rest/json/Quit()')
         self.vgwait(s3270)
 
     def test_s3270_japanese_930(self):
         self.s3270_japanese('cp930')
     def test_s3270_japanese_939(self):
         self.s3270_japanese('cp939')
+    def test_s3270_japanese_1390(self):
+        self.s3270_japanese('cp1390')
+    def test_s3270_japanese_1399(self):
+        self.s3270_japanese('cp1399')
+
+    def async_reply(self, p: playback):
+        '''Send '''
+        got = p.nread(39)
+        self.assertEqual('00000000007d5b7d115be40e8a82a3a589c1d0718d41d0619365a3a5ac97c9a19da5d2410fffef', got.hex())
+        p.send_records(1)
+
+    # s3270 Korean DBCS test.
+    def s3270_korean(self, codePage: str, show_codepage: str):
+
+        korean_text = "국민과함께하는민생토론회"
+
+        # Start playback.
+        pport, ts = unused_port()
+        with playback(self, 's3270/Test/korean.trc', port=pport) as p:
+            ts.close()
+
+            # Start s3270.
+            sport, ts = unused_port()
+            s3270 = Popen(vgwrap(['s3270', '-httpd', str(sport), '-utf8', '-codepage', codePage, f'127.0.0.1:{pport}']))
+            self.children.append(s3270)
+            self.check_listen(sport)
+            ts.close()
+
+            # Check the codepage config.
+            r = self.get(f'http://127.0.0.1:{sport}/3270/rest/json/Show(codepage)')
+            out = r.json()['result'][0]
+            self.assertEqual(out, show_codepage)
+
+            # Fill up the screen.
+            p.send_records(2)
+
+            # Prepare for the Enter.
+            reply_thread = threading.Thread(target=self.async_reply, args=[p])
+            reply_thread.start()
+
+            # Send the text.
+            self.get(f'http://127.0.0.1:{sport}/3270/rest/json/String("{korean_text}")')
+            self.get(f'http://127.0.0.1:{sport}/3270/rest/json/Enter())')
+
+        # Wait for the thread.
+        reply_thread.join(timeout=2)
+
+        # Wait for the process to exit successfully.
+        self.get(f'http://127.0.0.1:{sport}/3270/rest/json/Quit()')
+        self.vgwait(s3270)
+
+    def test_korean_933(self):
+        self.s3270_korean('korean', 'cp933 sbcs gcsgid 1173 cpgid 833 dbcs gcsgid 934 cpgid 834')
+    def test_korean_1364(self):
+        self.s3270_korean('korean-euro', 'cp1364 sbcs gcsgid 1173 cpgid 833 dbcs gcsgid 934 cpgid 834')
+
+    # s3270 PrintText() wrap test.
+    def s3270_dbcs_wrap(self, which: str):
+
+        japanese_text = "国内外の取材網を生かし国内外の取材網を生かし国内外の取材網を生かし国内外の取材網を生かし国内外の取材網を生かし"
+
+        # Start playback.
+        pport, ts = unused_port()
+        with playback(self, 's3270/Test/dbcs-wrap.trc', port=pport) as p:
+            ts.close()
+
+            # Start s3270.
+            sport, ts = unused_port()
+            s3270 = Popen(vgwrap(['s3270', '-httpd', str(sport), '-utf8', '-codepage', 'japanese-latin', f'127.0.0.1:{pport}']))
+            self.children.append(s3270)
+            self.check_listen(sport)
+            ts.close()
+
+            # Fill up the screen.
+            p.send_records(2)
+
+            # Send the text.
+            self.get(f'http://127.0.0.1:{sport}/3270/rest/json/String({japanese_text})')
+
+        # Get the resulting text from the display. The result should have the wrapped character displayed properly
+        # on the first line, and a space at the beginning of the second line.
+        if which == 'PrintText':
+            r = self.get(f'http://127.0.0.1:{sport}/3270/rest/json/PrintText(string)')
+            result = r.json()['result']
+            self.assertEqual('==>  ' + japanese_text[:38], result[21])
+            self.assertEqual(' ' + japanese_text[38:] + '                                             ', result[22])
+        elif which == 'Ascii1':
+            r = self.get(f'http://127.0.0.1:{sport}/3270/rest/json/Ascii1(22, 1, 2, 80)')
+            result = r.json()['result']
+            self.assertEqual('==>  ' + japanese_text[:38], result[0])
+            self.assertEqual(' ' + japanese_text[38:] + '                                             ', result[1])
+
+        # Wait for the process to exit successfully.
+        self.get(f'http://127.0.0.1:{sport}/3270/rest/json/Quit()')
+        self.vgwait(s3270)
+    # s3270 wrap tests.
+    def test_s3270_dbcs_wrap_printtext(self):
+        self.s3270_dbcs_wrap('PrintText')
+    def test_s3270_dbcs_wrap_ascii1(self):
+        self.s3270_dbcs_wrap('Ascii1')
 
 if __name__ == '__main__':
     unittest.main()
