@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2000-2022 Paul Mattes.
+ * Copyright (c) 2000-2025 Paul Mattes.
  * All rights reserved.
  * 
  * Redistribution and use in source and binary forms, with or without
@@ -34,6 +34,8 @@
  */
 
 #include "globals.h"
+
+#include <assert.h>
 #include <signal.h>
 #include "appres.h"
 #include "3270ds.h"
@@ -50,7 +52,7 @@
 #include "host.h"
 #include "keymap.h"
 #include "kybd.h"
-#include "lazya.h"
+#include "main_window.h"
 #include "names.h"
 #include "nvt.h"
 #include "popups.h"
@@ -62,6 +64,7 @@
 #include "task.h"
 #include "telnet.h"
 #include "trace.h"
+#include "txa.h"
 #include "unicodec.h"
 #include "utils.h"
 #include "varbuf.h"
@@ -167,7 +170,6 @@ bool screen_initted = true;
 bool escaped = true;
 bool isendwin = true;
 
-enum ts { TS_AUTO, TS_ON, TS_OFF };
 enum ts ab_mode = TS_AUTO;
 
 int windows_cp = 0;
@@ -218,7 +220,6 @@ static void status_printer(bool on);
 static int get_color_pair(int fg, int bg);
 static int color_from_fa(unsigned char fa);
 static void set_status_row(int screen_rows, int emulator_rows);
-static bool ts_value(const char *s, enum ts *tsp);
 static void relabel(bool ignored);
 static void init_user_colors(void);
 static void init_user_attribute_colors(void);
@@ -281,7 +282,7 @@ win32_perror_fatal(const char *fmt, ...)
     char *buf;
 
     va_start(ap, fmt);
-    buf = xs_vbuffer(fmt, ap);
+    buf = Vasprintf(fmt, ap);
     va_end(ap);
     win32_perror("%s", buf);
     x3270_exit(1);
@@ -380,6 +381,35 @@ model_cols(int m)
     }
 }
 
+/* Check the requested model against the console maximum. */
+static void
+check_console_max(void)
+{
+    int model_num_orig = model_num;
+    bool remodeled = false;
+
+    /* Switch models, if the chosen model won't fit on the maximum console window. */
+    while ((console_max.Y != 0 && console_max.Y < model_rows(model_num)) ||
+	   (console_max.X != 0 && console_max.X < model_cols(model_num))) {
+	if (model_num == 2) {
+	    /* Can't go any smaller. */
+	    popup_an_error("Maximum console window is %d rows x %d columns.\n"
+		    "wc3270 needs at least %d rows x %d columns. Try a smaller font.",
+		    console_max.Y, console_max.X,
+		    MODEL_2_ROWS, MODEL_2_COLS);
+	    x3270_exit(1);
+	}
+	set_rows_cols(model_num - 1, 0, 0);
+	remodeled = true;
+    }
+
+    if (remodeled) {
+	popup_an_error("Maximum console window is %d rows x %d columns.\n"
+		"Switching from model %d to model %d. Try a smaller font.",
+		console_max.Y, console_max.X, model_num_orig, model_num);
+    }
+}
+
 /*
  * Resize the newly-created console.
  *
@@ -407,6 +437,7 @@ resize_console(void)
     if (console_max.Y && want_bs.Y > console_max.Y) {
 	want_bs.Y = console_max.Y;
     }
+
     want_bs.X = model_cols(model_num);
     if (ov_cols > want_bs.X) {
 	want_bs.X = ov_cols;
@@ -471,34 +502,64 @@ resize_console(void)
 	 */
 	if (ov_cols > model_cols(model_num)) {
 	    if (ov_cols > console_cols) {
-		popup_an_error("Oversize columns (%d) truncated to maximum "
-			"window width (%d)", ov_cols, console_cols);
-			ov_cols = console_cols;
-			ov_changed = true;
+		ov_cols = console_cols;
+		popup_an_error("Oversize columns reduced to the maximum window width (%d).", ov_cols);
+		ov_changed = true;
 	    }
 	}
 
 	if (ov_rows > model_rows(model_num)) {
 	    if (ov_rows + XTRA_ROWS > console_rows) {
-		popup_an_error("Oversize rows (%d) truncated to maximum "
-			"window height (%d) - %d -> %d rows", ov_rows,
-			console_rows, XTRA_ROWS, console_rows - XTRA_ROWS);
 		ov_rows = console_rows - XTRA_ROWS;
-		if (ov_rows <= model_rows(model_num)) {
-		    ov_rows = 0;
+		if (ov_rows < model_rows(model_num)) {
+		    ov_rows = model_rows(model_num);
 		}
+		popup_an_error("Oversize rows reduced to the maximum window height (%d).", ov_rows);
 		ov_changed = true;
 	    }
 	}
     }
 
     if (ov_changed) {
+	popup_an_error("Try a smaller font.");
 	set_rows_cols(model_num, ov_cols, ov_rows);
     }
 
     return 0;
 }
 
+/* Console pre-initialization. */
+void
+screen_pre_init(void)
+{
+    /* Get the console handles. */
+    if (chandle == NULL) {
+	chandle = CreateFile("CONIN$", GENERIC_READ | GENERIC_WRITE,
+		FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
+	if (chandle == NULL) {
+	    win32_perror("CreateFile(CONIN$) failed");
+	    x3270_exit(1);
+	}
+    }
+
+    if (cohandle == NULL) {
+	cohandle = CreateFile("CONOUT$", GENERIC_READ | GENERIC_WRITE,
+		FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);
+	if (cohandle == NULL) {
+	    win32_perror("CreateFile(CONOUT$) failed");
+	    x3270_exit(1);
+	}
+    }
+
+    /* Get the console dimensions. */
+    if (GetConsoleScreenBufferInfo(cohandle, &base_info) == 0) {
+	win32_perror("GetConsoleScreenBufferInfo failed");
+	x3270_exit(1);
+    }
+    console_rows = base_info.srWindow.Bottom - base_info.srWindow.Top + 1;
+    console_cols = base_info.srWindow.Right - base_info.srWindow.Left + 1;
+}
+
 /*
  * Get a handle for the console.
  */
@@ -508,37 +569,17 @@ initscr(void)
     size_t buffer_size;
     CONSOLE_CURSOR_INFO cursor_info;
 
-    /* Get a handle to the console. */
-    chandle = CreateFile("CONIN$", GENERIC_READ | GENERIC_WRITE,
-	    FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
-    if (chandle == NULL) {
-	win32_perror("CreateFile(CONIN$) failed");
-	return NULL;
-    }
+    screen_pre_init();
     if (SetConsoleMode(chandle, ENABLE_PROCESSED_INPUT |
 				ENABLE_MOUSE_INPUT) == 0) {
 	win32_perror("SetConsoleMode failed");
 	return NULL;
     }
 
-    cohandle = CreateFile("CONOUT$", GENERIC_READ | GENERIC_WRITE,
-	    FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);
-    if (cohandle == NULL) {
-	win32_perror("CreateFile(CONOUT$) failed");
-	return NULL;
-    }
-
     console_window = get_console_hwnd();
+    set_main_window(console_window);
 
-    /* Get its dimensions. */
-    if (GetConsoleScreenBufferInfo(cohandle, &base_info) == 0) {
-	win32_perror("GetConsoleScreenBufferInfo failed");
-	return NULL;
-    }
-    console_rows = base_info.srWindow.Bottom - base_info.srWindow.Top + 1;
-    console_cols = base_info.srWindow.Right - base_info.srWindow.Left + 1;
-
-    /* Get its cursor configuration. */
+    /* Get the console cursor configuration. */
     if (GetConsoleCursorInfo(cohandle, &cursor_info) == 0) {
 	win32_perror("GetConsoleCursorInfo failed");
 	return NULL;
@@ -547,6 +588,9 @@ initscr(void)
     /* Get its maximum dimensions. */
     console_max = GetLargestConsoleWindowSize(cohandle);
 
+    /* Check the maximum size. */
+    check_console_max();
+
     /* Create the screen buffer. */
     sbuf = CreateConsoleScreenBuffer(GENERIC_READ | GENERIC_WRITE,
 	    FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, CONSOLE_TEXTMODE_BUFFER,
@@ -648,7 +692,7 @@ printw(char *fmt, ...)
     int i;
 
     va_start(ap, fmt);
-    buf = xs_vbuffer(fmt, ap);
+    buf = Vasprintf(fmt, ap);
     va_end(ap);
     sl = strlen(buf);
 
@@ -675,7 +719,7 @@ mvprintw(int row, int col, char *fmt, ...)
     cur_row = row;
     cur_col = col;
     va_start(ap, fmt);
-    buf = xs_vbuffer(fmt, ap);
+    buf = Vasprintf(fmt, ap);
     va_end(ap);
     sl = strlen(buf);
 
@@ -1232,10 +1276,6 @@ endwin(void)
 void
 screen_init(void)
 {
-    int want_ov_rows;
-    int want_ov_cols;
-    bool oversize = false;
-
     if (appres.interactive.menubar) {
 	menu_init();
     }
@@ -1245,57 +1285,9 @@ screen_init(void)
 	fprintf(stderr, "Can't initialize terminal.\n");
 	x3270_exit(1);
     }
-    want_ov_rows = ov_rows;
-    want_ov_cols = ov_cols;
     windows_cp = GetConsoleCP();
 
-    /*
-     * Respect the console size we are given.
-     */
-    while (console_rows < maxROWS || console_cols < maxCOLS) {
-	/*
-	 * First, cancel any oversize.  This will get us to the correct
-	 * model number, if there is any.
-	 */
-	if ((ov_cols && ov_cols > console_cols) ||
-	    (ov_rows && ov_rows > console_rows)) {
-
-	    ov_cols = 0;
-	    ov_rows = 0;
-	    oversize = true;
-	}
-
-	/* If we're at the smallest screen now, give up. */
-	if (model_num == 2) {
-	    fprintf(stderr, "Emulator won't fit on a %dx%d display.\n",
-		    console_rows, console_cols);
-	    x3270_exit(1);
-	}
-
-	/* Try a smaller model. */
-	set_rows_cols(model_num - 1, 0, 0);
-    }
-
-    /*
-     * Now, if they wanted an oversize, but didn't get it, try applying it
-     * again.
-     */
-    if (oversize) {
-	if (want_ov_rows > console_rows - 2) {
-	    want_ov_rows = console_rows - 2;
-	}
-	if (want_ov_rows < maxROWS) {
-	    want_ov_rows = maxROWS;
-	}
-	if (want_ov_cols > console_cols) {
-	    want_ov_cols = console_cols;
-	}
-	set_rows_cols(model_num, want_ov_cols, want_ov_rows);
-    }
-
-    /*
-     * Finally, if they want automatic oversize, see if that's possible.
-     */
+    /* If they want automatic oversize, see if that's possible. */
     if (ov_auto && (maxROWS < console_rows - 3 || maxCOLS < console_cols)) {
 	set_rows_cols(model_num, console_cols, console_rows - 3);
     }
@@ -1326,12 +1318,12 @@ screen_init(void)
 		ResAllBold, appres.c3270.all_bold);
     }
     if (ab_mode == TS_AUTO) {
-	ab_mode = mode.m3279? TS_ON: TS_OFF;
+	ab_mode = mode3279? TS_ON: TS_OFF;
     }
 
     /* If the want monochrome, assume they want green. */
     /* XXX: I believe that init_user_colors makes this a no-op. */
-    if (!mode.m3279) {
+    if (!mode3279) {
 	defattr |= FOREGROUND_GREEN;
 	xhattr |= FOREGROUND_GREEN;
 	if (ab_mode == TS_ON) {
@@ -1343,7 +1335,7 @@ screen_init(void)
     init_user_colors();
     init_user_attribute_colors();
 
-    if (mode.m3279) {
+    if (mode3279) {
 	oia_attr = cmap_fg[HOST_COLOR_GREY] | cmap_bg[HOST_COLOR_NEUTRAL_BLACK];
 	oia_bold_attr = oia_attr; /* not used */
 	oia_red_attr = FOREGROUND_RED | FOREGROUND_INTENSITY |
@@ -1403,29 +1395,6 @@ set_status_row(int screen_rows, int emulator_rows)
     }
 }
 
-/*
- * Parse a tri-state resource value.
- * Returns true for success, false for failure.
- */
-static bool
-ts_value(const char *s, enum ts *tsp)
-{
-    *tsp = TS_AUTO;
-
-    if (s != NULL && s[0]) {
-	size_t sl = strlen(s);
-
-	if (!strncasecmp(s, "true", sl)) {
-	    *tsp = TS_ON;
-	} else if (!strncasecmp(s, "false", sl)) {
-	    *tsp = TS_OFF;
-	} else if (strncasecmp(s, "auto", sl)) {
-	    return false;
-	}
-    }
-    return true;
-}
-
 /* Allocate a color pair. */
 static int
 get_color_pair(int fg, int bg)
@@ -1483,7 +1452,7 @@ init_user_attribute_colors(void)
 
 /*
  * Map a field attribute to a 3270 color index.
- * Applies only to m3279 mode -- does not work for mono.
+ * Applies only to 3279 mode -- does not work for mono.
  */
 static int
 color3270_from_fa(unsigned char fa)
@@ -1498,7 +1467,7 @@ color3270_from_fa(unsigned char fa)
 static int
 color_from_fa(unsigned char fa)
 {
-    if (mode.m3279) {
+    if (mode3279) {
 	int fg;
 
 	fg = color3270_from_fa(fa);
@@ -1605,7 +1574,7 @@ init_user_colors(void)
 	init_user_color(host_color[i].name, host_color[i].index);
     }
 
-    if (mode.m3279) {
+    if (mode3279) {
 	defattr = cmap_fg[HOST_COLOR_NEUTRAL_WHITE] |
 		  cmap_bg[HOST_COLOR_NEUTRAL_BLACK];
 	crosshair_color_init();
@@ -1667,7 +1636,7 @@ calc_attrs(int baddr, int fa_addr, int fa, bool *underlined,
     /* Compute the color. */
 
     /* Monochrome is easy, and so is color if nothing is specified. */
-    if (!mode.m3279 ||
+    if (!mode3279 ||
 	    (!ea_buf[baddr].fg &&
 	     !ea_buf[fa_addr].fg &&
 	     !ea_buf[baddr].bg &&
@@ -1708,7 +1677,7 @@ calc_attrs(int baddr, int fa_addr, int fa, bool *underlined,
     }
 
     if (!toggled(UNDERSCORE) &&
-	    mode.m3279 &&
+	    mode3279 &&
 	    (gr & (GR_BLINK | GR_UNDERLINE)) &&
 	    !(gr & GR_REVERSE) &&
 	    !bg) {
@@ -1716,7 +1685,7 @@ calc_attrs(int baddr, int fa_addr, int fa, bool *underlined,
 	a |= BACKGROUND_INTENSITY;
     }
 
-    if (!mode.m3279 &&
+    if (!mode3279 &&
 	    ((gr & GR_INTENSIFY) || (ab_mode == TS_ON) || FA_IS_HIGH(fa))) {
 
 	a |= FOREGROUND_INTENSITY;
@@ -2084,7 +2053,9 @@ screen_disp(bool erasing _is_unused)
 			int xaddr = baddr;
 
 			INC_BA(xaddr);
-			if (toggled(VISIBLE_CONTROL) &&
+			if (d == DBCS_LEFT_WRAP) {
+			    addch('.');
+			} else if (toggled(VISIBLE_CONTROL) &&
 				ea_buf[baddr].ec == EBC_null &&
 				ea_buf[xaddr].ec == EBC_null) {
 			    attrset(apply_select(cmap_fg[HOST_COLOR_NEUTRAL_BLACK] |
@@ -2140,6 +2111,8 @@ screen_disp(bool erasing _is_unused)
 			}
 			attrset(attr_this);
 			addch(blinkmap(blinking, underlined, u));
+		    } else if (d == DBCS_RIGHT_WRAP) {
+			addch('.');
 		    }
 		}
 	    }
@@ -2242,7 +2215,7 @@ decode_state(int state, bool limited, const char *skip)
 	vb_free(&r);
 	return "none";
     }
-    return lazya(vb_consume(&r));
+    return txdFree(vb_consume(&r));
 }
 
 /* Handle mouse events. */
@@ -2429,7 +2402,7 @@ decode_mflags(DWORD flags, decode_t names[])
     if (f != 0 && f != flags) {
 	vb_appendf(&r, "%s0x%x", any? "|": " ", f);
     }
-    return lazya(vb_consume(&r));
+    return txdFree(vb_consume(&r));
 }
 
 /* Redraw the screen in response to a screen resize event. */
@@ -2573,7 +2546,7 @@ trace_as_keymap(unsigned long xk, KEY_EVENT_RECORD *e)
     } else {
 	vb_appendf(&r, "<Key>%c", (unsigned char)xk);
     }
-    vtrace(" %s ->", lazya(vb_consume(&r)));
+    vtrace(" %s ->", txdFree(vb_consume(&r)));
 }
 
 /* Translate a Windows virtual key to a menubar abstract key. */
@@ -2706,14 +2679,14 @@ kybd_input2(INPUT_RECORD *ir)
 
     /* Catch PF keys. */
     if (k >= VK_F1 && k <= VK_F24) {
-	run_action(AnPF, IA_DEFAULT, lazyaf("%d", k - VK_F1 + 1), NULL);
+	run_action(AnPF, IA_DEFAULT, txAsprintf("%d", k - VK_F1 + 1), NULL);
 	return;
     }
 
     /* Then any other character. */
     if (ir->Event.KeyEvent.uChar.UnicodeChar) {
 	run_action(AnKey, IA_DEFAULT,
-		lazyaf("U+%04x", ir->Event.KeyEvent.uChar.UnicodeChar),
+		txAsprintf("U+%04x", ir->Event.KeyEvent.uChar.UnicodeChar),
 		NULL);
     } else {
 	vtrace(" dropped (no default)\n");
@@ -3139,18 +3112,19 @@ status_timing(struct timeval *t0, struct timeval *t1)
     static char no_time[] = ":??.?";
 
     if (t1->tv_sec - t0->tv_sec > (99*60)) {
-	strcpy(oia_timing, no_time);
+	strncpy(oia_timing, no_time, sizeof(oia_timing));
     } else {
 	unsigned long cs;	/* centiseconds */
 
 	cs = (t1->tv_sec - t0->tv_sec) * 10 +
 	     (t1->tv_usec - t0->tv_usec + 50000) / 100000;
 	if (cs < CM) {
-	    sprintf(oia_timing, ":%02ld.%ld", cs / 10, cs % 10);
+	    snprintf(oia_timing, sizeof(oia_timing), ":%02ld.%ld", cs / 10, cs % 10);
 	} else {
-	    sprintf(oia_timing, "%02ld:%02ld", cs / CM, (cs % CM) / 10);
+	    snprintf(oia_timing, sizeof(oia_timing), "%02ld:%02ld", cs / CM, (cs % CM) / 10);
 	}
     }
+    oia_timing[sizeof(oia_timing) - 1] = '\0';
 }
 
 void
@@ -3163,7 +3137,7 @@ void
 status_scrolled(int n)
 {
     if (n) {
-	Replace(scrolled_msg, xs_buffer("X Scrolled %d", n));
+	Replace(scrolled_msg, Asprintf("X Scrolled %d", n));
     } else {
 	Replace(scrolled_msg, NULL);
     }
@@ -3269,7 +3243,7 @@ draw_oia(void)
     }
 
     /* Offsets 0, 1, 2 */
-    if (mode.m3279) {
+    if (mode3279) {
 	attrset(cmap_fg[HOST_COLOR_NEUTRAL_BLACK] | cmap_bg[HOST_COLOR_GREY]);
     } else {
 	attrset(reverse_colors(defattr));
@@ -3324,7 +3298,7 @@ draw_oia(void)
 	    status_im? 'I': ' ',
 	    oia_printer? 'P': ' ');
     if (status_secure != SS_INSECURE) {
-	attrset(mode.m3279?
+	attrset(mode3279?
 		    (cmap_fg[(status_secure == SS_SECURE)?
 				HOST_COLOR_GREEN: HOST_COLOR_YELLOW] |
 		    cmap_bg[HOST_COLOR_NEUTRAL_BLACK]):
@@ -3509,7 +3483,7 @@ static void
 set_console_title(const char *text, bool selecting)
 {
     if (selecting) {
-	SetConsoleTitle(lazyaf("%s [select]", text));
+	SetConsoleTitle(txAsprintf("%s [select]", text));
     } else {
 	SetConsoleTitle(text);
     }
@@ -3550,7 +3524,7 @@ relabel(bool ignored _is_unused)
 	    hostname = reconnect_host;
 	}
 
-	screen_title(lazyaf("%s - wc3270", hostname));
+	screen_title(txAsprintf("%s - wc3270", hostname));
     } else {
 	screen_title("wc3270");
     }
@@ -3559,7 +3533,7 @@ relabel(bool ignored _is_unused)
 /**
  * Callback for changes in screen selection state.
  *
- * @param[in] selecting		true if selection in progress
+ * @param[in] now_selecting		true if selection in progress
  */
 static void
 screen_selecting_changed(bool now_selecting)
@@ -3621,10 +3595,11 @@ screen_wait_for_key(char *c)
 }
 
 /**
- * Query the screen for selections, wc3270 version.
- * This just defers over to the select logic.
+ * Check if an area of the screen is selected.
  *
  * @param[in] baddr	Buffer address.
+ *
+ * @return true if cell is selected
  */
 bool
 screen_selected(int baddr)
@@ -3658,7 +3633,7 @@ get_console_size(int *rows, int *cols)
 }
 
 /**
- * Stub for scrollbar function.
+ * Set the scrollbar thumb.
  *
  * @param[in] top	Where the top of the scrollbar should be (percentage)
  * @param[in] shown	How much of the scrollbar to show (percentage)
@@ -3672,6 +3647,19 @@ screen_set_thumb(float top _is_unused, float shown _is_unused,
 {
 }
 
+/**
+ * Change the model number, from a script.
+ *
+ * @param[in] mn        Model number
+ * @param[in] ovc       Oversize columns
+ * @param[in] ovr       Oversize rowa.
+ */
+void
+screen_change_model(int mn, int ovc, int ovr)
+{
+    assert(false);
+}
+
 /**
  * Enable or disable the cursor.
  *
@@ -3707,7 +3695,8 @@ static WORD color_attr[] = {
 void
 screen_color(pc_t pc)
 {
-    if (!SetConsoleTextAttribute(cohandle,
+    if (cohandle != NULL &&
+	    !SetConsoleTextAttribute(cohandle,
 		color_attr[pc]? color_attr[pc]: base_info.wAttributes)) {
 	win32_perror("Can't set console text attribute");
 	exit(1);
